# coding=utf-8

from ClickSQL.factor_table.factortable import FatctorTable

from test.conn import conn
Node = FatctorTable(conn, cik_dt='datetime', cik_iid='code')

if __name__ == '__main__':
    Node.add_factor('EDGAR_LOG.edgar_log', factor_names=['size'], cik_dt='date', cik_iid='ip')
    Node.add_factor('EDGAR_LOG.edgar_log', factor_names='cik', cik_dt='date', cik_iid='ip')
    # Node.add_factor('test.test2', factor_names='test2', cik_dt='cik_dt', cik_iid='cik_iid')
    # f = Node.fetch(reduced=True, add_limit=True)
    for df in Node:
        print(1)
        print(df)
    # print(f)
    pass



# coding=utf-8

import numpy as np
from functools import partial
from itertools import starmap
import pandas as pd


class SingleBrinson(object):
    """
    https://blog.csdn.net/qq_43382509/article/details/106035440
    """

    @staticmethod
    def cal_single_asi(rp: pd.Series, rb: pd.Series, wp: pd.Series, wb: pd.Series):
        """
        brinson 单期模型

        :param wp: portfolio weight
        :param rp: portfolio return matrix among single asset
        :param wb: benchmark weight
        :param rb: benchmark return matrix among single asset
        :return:
        """
        Rp = np.sum(np.dot(rp, wp))
        Rb = np.sum(np.dot(rb, wb))
        Re = Rp - Rb  # RE 为总超额收益

        Ra = np.sum(np.dot((wp - wb), rb))  # RTS 为择时收益
        Rs = np.sum(np.dot(wb, (rp - rb)))  # RSS 为择股收益
        Ri = np.sum(np.dot((wp - wb), (rp - rb)))  # RIE 为交互作用
        return Re, Ra, Rs, Ri

    pass


class MultiBrinson(object):
    """
    多期Brinson业绩归因分析——Carino模型
    https://zhuanlan.zhihu.com/p/42893026?utm_source=wechat_session
    """

    @staticmethod
    def cal_k(r, b):
        if r == b:
            return (np.log(1 + r) - np.log(1 + b)) / (r - b)
        else:
            return 1 / (1 + r)

    @classmethod
    def cal_kt(cls, rt, bt):
        return cls.cal_k(rt, bt)

    @staticmethod
    def cal_single(xt, kt, k, ):
        return (kt / k) * xt

    @classmethod
    def cal_a(cls, at, kt, k):
        return cls.cal_single(kt, k, at)

    @classmethod
    def cal_s(cls, st, kt, k):
        return cls.cal_single(kt, k, st)

    @classmethod
    def cal_i(cls, it, kt, k):
        return cls.cal_single(kt, k, it)

    @classmethod
    def cal_asi_base(cls, A_series, S_series, I_series, K_series, k):
        func = partial(cls.cal_single, k=k)
        A = sum(map(func, zip(A_series, K_series)))
        S = sum(map(func, zip(S_series, K_series)))
        I = sum(map(func, zip(I_series, K_series)))
        return A, S, I

    @classmethod
    def cal_asi_raw(cls, A_series, S_series, I_series, rt, bt, r, b):
        k = cls.cal_k(r, b)
        K_series = map(cls.cal_k, zip(rt, bt))
        return cls.cal_asi_base(cls, A_series, S_series, I_series, K_series, k)

    @classmethod
    def cal_asi(cls, wp: pd.Series, rp: pd.DataFrame, wb: pd.Series, rb: pd.DataFrame):
        Rp = np.dot(rp, wp)
        Rb = np.dot(rb, wb)
        r = np.prod(Rp + 1) - 1
        b = np.prod(Rb + 1) - 1
        func = partial(SingleBrinson.cal_single_asi, wp=wp, wb=wb)
        # tasks = [(p, b) for p, b in ]
        Re, A_series, S_series, I_series = zip(*list(starmap(func, zip(Rp, Rb))))
        return cls.cal_asi_raw(A_series, S_series, I_series, Rp, Rb, r, b)


if __name__ == '__main__':
    pass



# coding=utf-8
# coding=utf8
import responder

from ClickSQL.nodes.base import BaseSingleFactorBaseNode

conn_settings = 'clickhouse://default:Imsn0wfree@47.104.186.157:8123/system.columns'
# Tables =
api = responder.API()

# class GreetingResource:
#     def on_request(self, req, resp, *, greeting):   # or on_get...
#         resp.text = f"{greeting}, world!"
#         resp.headers.update({'X-Life': '42'})
#         resp.status_code = api.status_codes.HTTP_416
operator = BaseSingleFactorBaseNode(conn_settings, cols=[])

# @api.route("/{greeting}")
# class Service(object):
#     def __init__(self, conn_settings):
#         self.operator = BaseSingleFactorBaseNode(conn_settings, cols=[])


# def setup(self):
#     api.run(address=self.host, port=self.port, **self.api_kwargs)


# base_url = 'http://0.0.0.0:8279/'
# coding=utf8


# model_store_path = './static/model'


# @api.route("/")
# def hello_world(req, resp):
#     resp.text = "hello, world!"


FileExistsError_msg = 'data {} existed! please do not upload same data! '

#
# @api.route("/general/{sql}")
# def general(req, resp, *, sql):
#     pass


@api.route("/show_databases")
def show_datatbases(req, resp, ):
    res = operator('show databases').values.tolist()
    resp.status_code = 200
    resp.media = res


@api.route("/{db}/show_tables")
def db_table(req, resp, *, db):
    if db != 'query':
        res = operator(f'show tables from {db}').values.tolist()
        resp.status_code = 200
        resp.media = res
    else:
        resp.status_code = 400


@api.route("/{db}/{table}/show_columns")
def db_table(req, resp, *, db, table):
    res = operator(f'desc {db}.{table}').to_dict()
    resp.status_code = 200
    resp.media = res


@api.route("/query/{sql}")
def db_table(req, resp, *, sql):
    res = operator(sql).to_dict()
    resp.status_code = 200
    resp.media = res


if __name__ == '__main__':
    api.run(address='0.0.0.0', port=8237)
    pass



# coding=utf-8

from ClickSQL.nodes.base import BaseSingleFactorBaseNode
from collections import namedtuple, deque, Callable
import pandas as pd

CIK = namedtuple('CoreIndexKeys', ('dts', 'iid'))
CIKDATA = namedtuple('CoreIndexKeys', ('dts', 'iid'))
FactorInfo = namedtuple('FactorInfo', ('db_table', 'dts', 'iid', 'origin_factor_names', 'alias', 'sql', 'conditions'))


def SmartDataFrame(df: pd.DataFrame, db_table: str, dts: str, iid: str, origin_factor_names: str, alias: str, sql: str,
                   conditions: str):
    sproperty = {'_db_table': property(lambda x: db_table),
                 '_cik_dt': property(lambda x: dts),
                 '_cik_iid': property(lambda x: iid),
                 '_origin_factor_names': property(lambda x: origin_factor_names),
                 '_alias': property(lambda x: alias),
                 '_sql': property(lambda x: sql),
                 '_conditions': property(lambda x: conditions),

                 }

    result_cls = type('SmartDataFrame', (pd.DataFrame,), sproperty)
    return result_cls(df)


class FactorCheckHelper(object):
    @staticmethod
    def check_alias(factor_names: (list,), as_alias: (list, tuple, str) = None):
        if as_alias is None:
            alias = len(factor_names) * [None]
        elif isinstance(as_alias, str):
            alias = [as_alias]
        elif isinstance(as_alias, (list, tuple)):
            if len(as_alias) != len(factor_names):
                raise ValueError('as_alias is not match factor_names')
            else:
                alias = as_alias
        else:
            raise ValueError('alias only accept list tuple str!')
        return alias

    @staticmethod
    def check_factor_names(factor_names: (list, tuple, str)):
        if isinstance(factor_names, str):
            factor_names = [factor_names]
        elif isinstance(factor_names, (list, tuple)):
            factor_names = list(factor_names)
        else:
            raise ValueError('columns only accept list tuple str!')
        return factor_names

    @staticmethod
    def check_cik_dt(cik_dt, default_cik_dt):
        if cik_dt is not None:
            pass

        elif cik_dt is None:
            cik_dt = default_cik_dt
        else:
            raise NotImplementedError('cik_dt is not setup!')
        return cik_dt

    @staticmethod
    def check_cik_iid(cik_iid, default_cik_iid):
        if cik_iid is not None:
            pass

        elif cik_iid is None:
            cik_iid = default_cik_iid
        else:
            raise NotImplementedError('cik_dt is not setup!')
        return cik_iid


class _Factors(deque):

    @staticmethod
    def _get_factor_without_check(db_table, factor_names: (list, tuple, str), cik_dt=None, cik_iid=None,
                                  conds: str = '1', as_alias: (list, tuple, str) = None):
        """

        :param db_table:
        :param factor_names:
        :param cik_dt:
        :param cik_iid:
        :param conds:  conds = @test1>1 | @test2<1
        :return:
        """
        factor_names = FactorCheckHelper.check_factor_names(factor_names)
        alias = FactorCheckHelper.check_alias(factor_names, as_alias=as_alias)
        # rename variables
        f_names_list = [f if (a is None) or (f == a) else f"{f} as {a}" for f, a in zip(factor_names, alias)]
        cols_str = ','.join(f_names_list)

        conditions = '1' if conds == '1' else conds.replace('&', 'and').replace('|', 'or').replace('@', '')
        cik_dt_str = f"{cik_dt} as cik_dt" if cik_dt != 'cik_dt' else cik_dt
        cik_iid_str = f"{cik_iid} as cik_iid" if cik_iid != 'cik_iid' else cik_iid

        sql = f'select {cols_str}, {cik_dt_str}, {cik_iid_str}  from {db_table} where {conditions}'

        return FactorInfo(db_table, cik_dt, cik_iid, ','.join(map(str, factor_names)), ','.join(map(str, alias)),
                          sql, conds)  #

    def show_factors(self, reduced=False, to_df=True):
        if reduced:
            # ('db_table', 'dts', 'iid', 'origin_factor_names', 'alias', 'sql', 'conditions')
            # ['db_table', 'dts', 'iid', 'conditions']
            cols = list(FactorInfo._fields[:3]) + [FactorInfo._fields[-1]]

            f = pd.DataFrame(self, columns=FactorInfo._fields)
            factor_name_col = FactorInfo._fields[3]
            alias_col = FactorInfo._fields[4]

            # can_merged_index = (fgroupby['sql'].count() > 1).reset_index()
            # can_merged_index = can_merged_index[can_merged_index['sql']]
            # can_merged_index = fgroupby.count().index
            factors = []
            for (db_table, dts, iid, conditions), df in f.groupby(cols):
                # masks = (f['db_table'] == db_table) & (f['dts'] == dts) & (f['iid'] == iid) & (
                #         f['conditions'] == conditions)
                cc = df[[factor_name_col, alias_col]].apply(lambda x: ','.join(x))
                origin_factor_names = cc[factor_name_col].split(',')
                alias = cc[alias_col].split(',')
                origin_factor_names_new, alias_new = zip(*list(set(zip(origin_factor_names, alias))))
                alias_new = list(map(lambda x: x if x != 'None' else None, alias_new))

                # cik_dt, cik_iid = self.check_cik_dt(cik_dt=dts, default_cik_dt=self._cik.dts), self.check_cik_iid(
                #     cik_iid=iid, default_cik_iid=self._cik.iid)
                # add_factor process have checked
                res = self._get_factor_without_check(db_table, origin_factor_names_new, cik_dt=dts, cik_iid=iid,
                                                     conds=conditions, as_alias=alias_new)
                factors.append(res)

        else:
            factors = self
        if to_df:
            return pd.DataFrame(factors, columns=FactorInfo._fields)
        else:
            return factors

    def fetch_iter(self, query, filter_cond_dts, filter_cond__ids, reduced=True, add_limit=False):
        if not isinstance(query, Callable):
            raise ValueError('query must database connector with __call__')

        factors = self.show_factors(reduced=reduced, to_df=False)

        for db_table, dts, iid, origin_factor_names, alias, sql, conditions in factors:
            ## todo 可能存在性能点
            if add_limit:
                sql2 = f"select * from ({sql}) where {filter_cond_dts} and {filter_cond__ids} limit 100"
            else:
                sql2 = f"select * from ({sql}) where {filter_cond_dts} and {filter_cond__ids}"

            df = query(sql2)
            res = SmartDataFrame(df, db_table, dts, iid, origin_factor_names, alias, sql, conditions).set_index(
                ['cik_dt', 'cik_iid'])

            # ['db_table', 'dts', 'iid', 'origin_factor_names', 'alias', 'sql', 'conditions']
            yield res

    def fetch_sql(self, query, filter_cond_dts, filter_cond__ids, reduced=True, add_limit=False):
        if not isinstance(query, Callable):
            raise ValueError('query must database connector with __call__')
        factors = self.show_factors(reduced=reduced, to_df=False)
        sql_list = []
        for db_table, dts, iid, origin_factor_names, alias, sql, conditions in factors:
            ## todo 可能存在性能点
            if add_limit:
                sql2 = f"select * from ({sql}) where {filter_cond_dts} and {filter_cond__ids} limit 100"
            else:
                sql2 = f"select * from ({sql}) where {filter_cond_dts} and {filter_cond__ids}"
            sql_list.append(sql2)

        from functools import reduce

        def join(sql1, sql2):
            settings = ' settings join'
            sql = f"select * from ({sql1}) all full join ({sql2}) using (cik_dt,cik_iid)  {settings}"
            return sql

        s = reduce(lambda x, y: join(x, y), sql_list)
        df = query(s)
        res = SmartDataFrame(df, db_table, dts, iid, origin_factor_names, alias, sql, conditions).set_index(
            ['cik_dt', 'cik_iid'])

        # ['db_table', 'dts', 'iid', 'origin_factor_names', 'alias', 'sql', 'conditions']
        yield res


class FatctorTable(FactorCheckHelper):
    __Name__ = "基础因子库单因子表"

    def __init__(self, *args, **kwargs):
        # super(FatctorTable, self).__init__(*args, **kwargs)
        self._node = BaseSingleFactorBaseNode(*args, **kwargs)

        cik_dt = None if 'cik_dt' not in kwargs.keys() else kwargs['cik_dt']
        cik_iid = None if 'cik_iid' not in kwargs.keys() else kwargs['cik_iid']
        self._cik = CIK(cik_dt, cik_iid)
        self._cik_data = None
        self._checked = False
        self.__auto_check_cik__()
        self._factors = _Factors()

        self._strict_cik = True if 'strict_cik' not in kwargs.keys() else kwargs['strict_cik']
        # self.append = self.add_factor

        self._cik_dts = None
        self._cik_iids = None

    def __auto_check_cik__(self):
        if not self._checked and (self._cik.dts is None or self._cik.iid is None):
            raise NotImplementedError('cik(dts or iid) is not setup!')
        else:
            self._checked = True

    # def __check_cik__(self, cik_dt=None, cik_iid=None):
    #     if cik_dt is not None:
    #         pass
    #
    #     elif cik_dt is None:
    #         cik_dt = self._cik.dts
    #     else:
    #         raise NotImplementedError('cik_dt is not setup!')
    #
    #     if cik_iid is not None:
    #         pass
    #     elif cik_iid is None :
    #         cik_iid = self._cik.iid
    #     else:
    #         raise NotImplementedError('cik_iid is not setup!')
    #     return cik_dt, cik_iid

    def setup_cik(self, cik_dt_col: str, cik_iid_col: str):
        """
        设置 cik 列名
        :param cik_dt_col:
        :param cik_iid_col:
        :return:
        """

        self._cik = CIK(cik_dt_col, cik_iid_col)

    # def getDB(self, db):
    #     self.db = db

    def add_factor(self, db_table, factor_names: (list, tuple, str), cik_dt=None, cik_iid=None,
                   as_alias: (list, tuple, str) = None):
        conds = '1'  # not allow to set conds
        cik_dt, cik_iid = self.check_cik_dt(cik_dt=cik_dt, default_cik_dt=self._cik.dts), self.check_cik_iid(
            cik_iid=cik_iid, default_cik_iid=self._cik.iid)
        res = self._factors._get_factor_without_check(db_table, factor_names, cik_dt=cik_dt, cik_iid=cik_iid,
                                                      conds=conds, as_alias=as_alias)
        self._factors.append(res)

    def show_factors(self, reduced=False, to_df=True):
        return self._factors.show_factors(reduced=reduced, to_df=to_df)

        # no_duplicates_df = f.eval("+".join(cols))
        ## todo auto merge same condition,dbtable,dts,iid
        # return can_merged_index

    def __iter__(self):
        return self._factors.fetch_iter(self._node, self.cik_dt, self.cik_iid, reduced=True,
                                        add_limit=False)

    def head(self, reduced=True, ):
        """
        quick look top data
        :param reduced:
        :return:
        """

        return self.fetch(self, reduced=reduced, add_limit=True)

    def fetch(self, reduced=True, add_limit=False):
        if self._strict_cik:
            if self._cik_dts is None:
                raise KeyError('cik_dts is not setup!')
            if self._cik_iids is None:
                raise KeyError('cik_iids is not setup!')

        return pd.concat(
            self._factors.fetch_iter(self._node, self.cik_dt, self.cik_iid, reduced=reduced,
                                     add_limit=add_limit), axis=1)

    @property
    def cik_dt(self):
        dt_format = "%Y%m%d"
        if self._cik_dts is None:
            return "  1 "

        else:
            cik_dts_str = "','".join(map(lambda x: x.strftime(dt_format), pd.to_datetime(self._cik_dts)))
            return f" cik_dt in ('{cik_dts_str}') "

    @cik_dt.setter
    def set_cik_dt(self, cik_dt: list):
        self._cik_dts = cik_dt

    @property
    def cik_iid(self):
        if self._cik_iids is None:
            return "  1 "

        else:
            cik_iid_str = "','".join(map(lambda x: x, self._cik_iids))
            return f" cik_iid in ('{cik_iid_str}') "

    @cik_iid.setter
    def set_cik_iid(self, cik_iid: list):
        self._cik_iids = cik_iid
        pass

    # def where(self, cik_dt: list = None, cik_iid: list = None, cik_dt_format="%Y%m%d"):
    #     extra_conds = []
    #     if cik_dt is None:
    #         pass
    #     else:
    #         cik_dt_ = ','.join(pd.to_datetime(cik_dt).strftime(cik_dt_format))
    #         cik_dt_cond = f"cik_dt in ({cik_dt_}) "
    #
    #     if cik_iid is None:
    #         pass


if __name__ == '__main__':
    FatctorTable()
    pass



# coding=utf-8


class ParameterKeyError(Exception):
    pass


class ParameterTypeError(Exception):
    pass


class DatabaseTypeError(Exception):
    pass


class DatabaseNotExists(Exception): pass


class DatabaseError(Exception):
    pass


class HeartbeatCheckFailure(Exception):
    pass


class ClickHouseTableExistsError(Exception):
    pass


class ClickHouseTableNotExistsError(Exception):
    pass


class ArgumentError(Exception):
    """Raised when an invalid or conflicting function argument is supplied.

    This error generally corresponds to construction time state errors.

    """
    pass



# coding=utf-8
import pandas as pd
import numpy as np
import warnings
from scipy.stats.mstats import winsorize
from scipy import linalg


def orth(df, cols=None, index=None):
    cols = df.columns if cols is None else cols
    index = df.index if index is None else index
    a2 = df.value

    orthed_a = np.array(linalg.orth(a2, rcond=0), dtype=float)
    if orthed_a.shape == a2.shape:
        return pd.DataFrame(orthed_a, index=index, columns=cols)
    else:
        raise ValueError('exists highly related factors, please check data!')


def detect_series_na(series, raise_error=True):
    na_value = sum(series.isna())
    if na_value >= 1:
        msg = f'own {na_value}  NA value '
        if raise_error:
            raise ValueError(msg)
        else:
            warnings.warn(msg)
            return na_value
    else:
        return None


def _check_series_type(series):
    if isinstance(series, pd.Series):
        pass
    else:
        series = pd.Series(series)
    return series


class Outlier(object):
    """
    provide outlier detector

    """
    __slots__ = []

    @staticmethod
    def box(series, time=1.5):
        series = _check_series_type(series)
        L, U = np.percentile(sorted(series), [25, 75])
        IQR = U - L
        return np.clip(series, L - time * IQR, U + time * IQR)

    @staticmethod
    def winsorize(series, limits=[0.1, 0.2]):
        series = _check_series_type(series)
        return pd.Series(winsorize(series, limits=limits))

    @staticmethod
    def percentile(series, threshold=0.95):
        """

        :param series:
        :param threshold:
        :return:
        """
        series = _check_series_type(series)
        diff = threshold / 2.0
        lower, upper = np.percentile(sorted(series), [diff * 100, 100 - diff * 100])
        return np.clip(series, lower, upper)

    @staticmethod
    def mad(series, time=3):
        """
        median absolute deviation

        :param series:
        :param time:
        :return:
        """
        # if isinstance(series, pd.Series):
        #     pass
        # else:
        #     series = pd.Series(series)
        series = _check_series_type(series)
        median = series.quantile(0.5)
        new_median = (series - median).abs().quantile(0.5)
        upper = median + new_median * time
        lower = median - new_median * time
        return np.clip(series, lower, upper)

    @staticmethod
    def sigma(series, time=3):
        # if isinstance(series, pd.Series):
        #     pass
        # else:
        #     series = pd.Series(series)
        series = _check_series_type(series)
        std = series.std()
        mean = series.mean()

        upper = mean + std * time
        lower = mean - std * time
        return np.clip(series, lower, upper)


class Scaling(object):
    """
    squered_transform
    equal
    sqrtroot
    log
    recipocal
    boxcox
    signlog
    sigmoid

    """
    __slots__ = []

    @staticmethod
    def sigmoid(series):
        series = _check_series_type(series)
        return 1 / (1 + np.exp(-1 * series))

    @staticmethod
    def signlog(series):
        series = _check_series_type(series)
        return np.sign(series) * np.log(np.abs(series))

    @staticmethod
    def fractional(series):
        series = _check_series_type(series)
        return series / 10 ** np.ceil(np.log10(series.abs().max()))

    @staticmethod
    def range(series, range_min=0, range_max=1):
        series = _check_series_type(series)
        if range_max >= range_min:
            max = np.max(series)
            min = np.min(series)
            max_min = max - min
            if max_min > 0:
                return (series - min) / max_min * (range_max - range_min) + range_min
            else:
                raise ValueError('series got equally data or nan value')
        else:
            raise ValueError('range_max should be greater than range_min')

    @staticmethod
    def z_score(series):
        """

        :param series:
        :return:
        """
        series = _check_series_type(series)
        mean = np.mean(series)
        std = np.std(series)
        return (series - mean) / std

    @staticmethod
    def rank_z_score(series):
        """

        :param series:
        :return:
        """
        series = _check_series_type(series).rank()
        mean = np.mean(series)
        std = np.std(series)
        return (series - mean) / std

    @staticmethod
    def max_diff(series, ):
        """
        极差=最大值-最小值
        :param series:
        :return:
        """
        series = _check_series_type(series)
        dmax, dmin = np.max(series), np.min(series)
        max_diff = dmax - dmin
        return (series - dmin) / max_diff

    @staticmethod
    def cauterise(series):
        return series - series.mean()


if __name__ == '__main__':
    c = Scaling.signlog(np.random.random(size=(100,)))
    print(1)
    pass
